require 'minitest/autorun'
require 'att/swift'

class TestATTSwift < MiniTest::Unit::TestCase

  def setup
    @auth_uri = URI 'http://example/auth/'
    @store_uri = URI 'http://storage.example/'

    @swift = ATT::Swift.new @auth_uri, 'username', 'key'
    @swift.http = @http = FakeHTTP.new

    @http.expected_response(200, nil,
                            'X-Auth-Token' => 'token',
                            'X-Storage-Url' => 'http://storage.example/')
  end

  def test_initialize
    swift = ATT::Swift.new @auth_uri.to_s, 'username', 'key'
    swift.http = @http

    swift.authenticate

    assert_equal 'token', swift.auth_token
  end

  def test_authenticate
    @swift.authenticate

    assert_equal 'token', @swift.auth_token
    assert_equal URI('http://storage.example/'), @swift.storage_uri
    assert_equal 'token', @http.override_headers['X-Auth-Token']

    uri, req = @http.requests.shift

    assert_equal @auth_uri + 'v1.0', uri

    assert_equal '/auth/v1.0', req.path
    assert_equal %w[username], req.to_hash['x-auth-user']
    assert_equal %w[key],      req.to_hash['x-auth-key']

    @swift.authenticate

    assert_empty @http.requests, 'authenticate did not cache the auth token'
  end

  def test_chunk_objects
    expected_body_1 = [
      { 'name' => 'hello', 'hash' => '5d41402abc4b2a76b9719d911017c592',
        'bytes' => 5 }
    ]

    expected_body_2 = [
      { 'name' => 'world', 'hash' => '7d793037a0760186574b0282f2f435e7',
        'bytes' => 5 },
    ]

    @http.expected_response 200, JSON.dump(expected_body_1)
    @http.expected_response 200, JSON.dump(expected_body_2)
    @http.expected_response 200, JSON.dump([])

    chunks = []

    @swift.chunk_objects 'container', nil, 1 do |objects|
      chunks << objects
    end

    assert_equal [expected_body_1, expected_body_2], chunks

    @http.requests.shift # auth

    uri, = @http.requests.shift
    assert_equal @store_uri + 'container?format=json&limit=1', uri

    uri, = @http.requests.shift
    assert_equal @store_uri + 'container?format=json&marker=hello&limit=1', uri

    uri, = @http.requests.shift
    assert_equal @store_uri + 'container?format=json&marker=world&limit=1', uri
  end

  def test_containers
    expected_body = [
      { 'name' => 'a', 'count' => 12, 'bytes' => 345 },
    ]

    @http.expected_response 200, JSON.dump(expected_body)

    containers = @swift.containers

    assert_equal expected_body, containers

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + '?format=json', uri
  end

  def test_containers_limit
    expected_body = [
      { 'name' => 'b', 'count' => 12, 'bytes' => 345 },
    ]

    @http.expected_response 200, JSON.dump(expected_body)

    @swift.containers nil, 2

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + '?format=json&limit=2', uri
  end

  def test_containers_marker
    expected_body = [
      { 'name' => 'b', 'count' => 12, 'bytes' => 345 },
    ]

    @http.expected_response 200, JSON.dump(expected_body)

    @swift.containers 'a'

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + '?format=json&marker=a', uri
  end

  def test_copy_object
    @http.expected_response 201

    assert @swift.copy_object('src_container', 'src_object',
                              'dst_container', 'dst_object')

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'dst_container/dst_object', uri

    assert_kind_of Net::HTTP::Put, req
    assert_equal '/src_container/src_object', req['x-copy-from']
  end

  def test_create_container
    @http.expected_response 201

    assert @swift.create_container 'test'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'test', uri

    assert_kind_of Net::HTTP::Put, req
  end

  def test_create_container_exists
    @http.expected_response 200

    refute @swift.create_container 'test'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'test', uri

    assert_kind_of Net::HTTP::Put, req
  end

  def test_delete_container
    @http.expected_response 204

    assert @swift.delete_container 'test'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'test', uri

    assert_kind_of Net::HTTP::Delete, req
  end

  def test_delete_container_nonexistent
    @http.expected_response 404

    refute @swift.delete_container 'test'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'test', uri

    assert_kind_of Net::HTTP::Delete, req
  end

  def test_delete_container_not_empty
    @http.expected_response 409

    e = assert_raises ATT::Swift::Error do
      @swift.delete_container 'test'
    end

    assert_equal 'container test is not empty - 409', e.message

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'test', uri

    assert_kind_of Net::HTTP::Delete, req
  end

  def test_delete_object
    @http.expected_response 204

    assert @swift.delete_object 'container', 'object'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri

    assert_kind_of Net::HTTP::Delete, req
  end

  def test_delete_object_nonexistent
    @http.expected_response 404

    refute @swift.delete_object 'container', 'object'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri

    assert_kind_of Net::HTTP::Delete, req
  end

  def test_objects
    expected_body = [
      { 'name' => 'hello', 'hash' => '5d41402abc4b2a76b9719d911017c592',
        'bytes' => 5 },
      { 'name' => 'world', 'hash' => '7d793037a0760186574b0282f2f435e7',
        'bytes' => 5 },
    ]

    @http.expected_response 200, JSON.dump(expected_body)

    objects = @swift.objects 'container'

    assert_equal expected_body, objects

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + 'container?format=json', uri
  end

  def test_objects_limit
    expected_body = [
      { 'name' => 'hello', 'hash' => '5d41402abc4b2a76b9719d911017c592',
        'bytes' => 5 },
    ]

    @http.expected_response 200, JSON.dump(expected_body)

    objects = @swift.objects 'container', nil, 2

    assert_equal expected_body, objects

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + 'container?format=json&limit=2', uri
  end

  def test_objects_marker
    expected_body = [
      { 'name' => 'hello', 'hash' => '5d41402abc4b2a76b9719d911017c592',
        'bytes' => 5 },
    ]

    @http.expected_response 200, JSON.dump(expected_body)

    objects = @swift.objects 'container', 'h'

    assert_equal expected_body, objects

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + 'container?format=json&marker=h', uri
  end

  def test_object_info
    @http.expected_response(200, nil,
                            'Content-Length' => '3072',
                            'Content-Type' => 'application/octet-stream',
                            'ETag' => 'a2a5648d09a83c6a85ddd62e4a22309c',
                            'Last-Modified' => 'Wed, 22 Aug 2012 22:51:28 GMT',
                            'X-Object-Meta-Meat' => 'Bacon')

    info = @swift.object_info 'container', 'object'

    expected = {
      'metadata' => {
        'meat' => 'Bacon',
      },

      'content-length' => '3072',
      'content-type'   => 'application/octet-stream',
      'etag'           => 'a2a5648d09a83c6a85ddd62e4a22309c',
      'last-modified'  => 'Wed, 22 Aug 2012 22:51:28 GMT',
    }

    assert_equal expected, info

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Head, req
  end

  def test_object_info_missing
    @http.expected_response 404

    info = @swift.object_info 'container', 'object'

    assert_nil info

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Head, req
  end

  def test_object_metadata
    @http.expected_response(200, nil,
                            'Content-Length' => '3072',
                            'Content-Type' => 'application/octet-stream',
                            'ETag' => 'a2a5648d09a83c6a85ddd62e4a22309c',
                            'Last-Modified' => 'Wed, 22 Aug 2012 22:51:28 GMT',
                            'X-Object-Meta-Meat' => 'Bacon')

    metadata = @swift.object_metadata 'container', 'object'

    expected = {
      'meat' => 'Bacon',
    }

    assert_equal expected, metadata

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Head, req
  end

  def test_object_metadata_missing
    @http.expected_response 404

    metadata = @swift.object_metadata 'container', 'object'

    assert_nil metadata

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Head, req
  end

  def test_paginate_objects
    expected_body_1 = [
      { 'name' => 'hello', 'hash' => '5d41402abc4b2a76b9719d911017c592',
        'bytes' => 5 }
    ]

    expected_body_2 = [
      { 'name' => 'world', 'hash' => '7d793037a0760186574b0282f2f435e7',
        'bytes' => 5 },
    ]

    @http.expected_response 200, JSON.dump(expected_body_1)
    @http.expected_response 200, JSON.dump(expected_body_2)
    @http.expected_response 200, JSON.dump([])

    paginated = []

    @swift.paginate_objects 'container', nil, 1 do |object|
      paginated << object
    end

    assert_equal [expected_body_1, expected_body_2].flatten, paginated

    @http.requests.shift # auth

    uri, = @http.requests.shift
    assert_equal @store_uri + 'container?format=json&limit=1', uri

    uri, = @http.requests.shift
    assert_equal @store_uri + 'container?format=json&marker=hello&limit=1', uri

    uri, = @http.requests.shift
    assert_equal @store_uri + 'container?format=json&marker=world&limit=1', uri
  end

  def test_read_object
    @http.expected_response(200, 'hello',
                            'ETag' => Digest::MD5.hexdigest('hello'))

    object = @swift.read_object 'container', 'object'

    assert_equal 'hello', object

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
  end

  def test_read_object_not_found
    @http.expected_response 404

    e = assert_raises ATT::Swift::Error do
      @swift.read_object 'container', 'object'
    end

    assert_equal 'object object in container not found - 404', e.message

    @http.requests.shift # auth

    uri, = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
  end

  def test_set_object_metadata
    @http.expected_response 202

    @swift.set_object_metadata 'container', 'object', 'meat' => 'bacon'

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Post, req

    expected = {
      'x-object-meta-meat' => %w[bacon]
    }

    assert_equal expected,
                 req.to_hash.select { |key,| key =~ /^x-object-meta-/ }
  end

  def test_write_object
    digest = Digest::MD5.hexdigest 'hello'
    @http.expected_response 201, nil, 'ETag' => digest

    md5 = @swift.write_object 'container', 'object', 'hello'

    assert_equal digest, md5

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Put, req
    assert_equal 'hello', req.body
  end

  def test_write_object_block
    digest = Digest::MD5.hexdigest 'hello'
    @http.expected_response 201, nil, 'ETag' => digest

    md5 = @swift.write_object 'container', 'object' do |io|
      io.write 'hello'
    end

    assert_equal digest, md5

    @http.requests.shift # auth

    uri, req = @http.requests.shift

    assert_equal @store_uri + 'container/object', uri
    assert_kind_of Net::HTTP::Put, req
    assert_kind_of IO, req.body_stream
  end

  class FakeHTTP
    attr_accessor :requests, :responses, :override_headers

    def initialize
      @requests  = []
      @responses = []

      @override_headers = {}
    end

    def debug_output= ignored
    end

    def expected_response code, body = nil, headers = {}
      code     = code.to_s
      klass    = Net::HTTPResponse::CODE_TO_OBJ[code]

      response = klass.new '1.1', code, ''
      response.initialize_http_header headers

      @override_headers.each do |name, value|
        response[name] = value
      end

      if body then
        response.instance_variable_set :@read, true
        response.instance_variable_set :@body, body
      end

      @responses << response
    end

    def request uri, req = nil
      @requests << [uri, req]

      raise "response list empty for #{uri}" if @responses.empty?

      response = @responses.shift

      if block_given? then
        yield response
      else
        response
      end
    end
  end

end

